<?php
require_once($INC . "/prof.inc.php");
require_once($INC . "/log.inc.php");

set_error_handler("rg_error_handler");
register_shutdown_function("rg_error_shutdown");

$rg_util_error = "";

function rg_util_set_error($str)
{
	global $rg_util_error;
	$rg_util_error = $str;
}

function rg_util_error()
{
	global $rg_util_error;
	return $rg_util_error;
}

/*
 * This array will keep all registered function for templates
 */
$rg_template_functions = array();

/*
 * Register a function to be called when a @@func:func_name:var@@ construction
 * is found in a template.
 * Please note that the function is called once per template.
 */
function rg_template_func($name, $real_func_name)
{
	global $rg_template_functions;

	$rg_template_functions[$name] = $real_func_name;
}

function sha512($m)
{
	return hash('sha512', $m);
}

function rg_1024($v)
{
	if ($v <= 9999)
		return number_format($v);
	$v /= 1024;

	if ($v <= 9999)
		return number_format($v) . "KiB";
	$v /= 1024;

	if ($v <= 9999)
		return number_format($v) . "MiB";
	$v /= 1024;

	if ($v <= 9999)
		return number_format($v) . "GiB";
	$v /= 1024;

	return number_format($v) . "TiB";
}

/*
 * Unique ID generator
 */
function rg_id($len)
{
	rg_prof_start("urandom");

	$id = "";

	$f = @fopen("/dev/urandom", "r");
	if ($f !== NULL) {
		$buf = @fread($f, 128);
		if ($buf !== NULL)
			$id = sha512($buf);
		fclose($f);
	}

	if (empty($id))
		$id = sha1(mt_rand() . serialize($_REQUEST));

	rg_prof_end("urandom");
	return substr($id, 0, $len);
}

/*
 * Locks a file
 */
$_lock = array();
function rg_lock($file)
{
	global $_lock;
	global $rg_lock_dir;

	if (!isset($rg_lock_dir))
		$rg_lock_dir = "/var/lib/rocketgit/locks";

	// Double locking?
	if (isset($_lock[$file])) {
		rg_internal_error("Double locking [$file]: "
			. rg_array2string($_lock));
		return FALSE;
	}

	$f = @fopen($rg_lock_dir . "/" . $file, "w");
	if ($f === FALSE) {
		rg_internal_error("Cannot open lock $file ($php_errormsg).");
		return FALSE;
	}

	if (!flock($f, LOCK_EX | LOCK_NB)) {
		fclose($f);
		return FALSE;
	}

	fwrite($f, getmypid() . "\n");

	$_lock[$file] = $f;

	return TRUE;
}

function rg_lock_or_exit($file)
{
	if (rg_lock($file) === FALSE)
		exit(0);
}

function rg_unlock($file)
{
	global $_lock;

	if (isset($_lock[$file]))
		fclose($_lock[$file]);
}

function rg_load()
{
	return intval(file_get_contents("/proc/loadavg"));
}

/*
 * Outputs a string to browser, XSS safe
 * Thanks OWASP!
 */
function rg_xss_safe($str)
{
	return htmlspecialchars($str, ENT_QUOTES | ENT_HTML401, 'UTF-8');
}

/*
 * Builds URLs
 */
function rg_re_url($area)
{
	if (isset($_REQUEST['rwe']))
		return $area;

	$add = "";
	if (isset($_SERVER['PHP_SELF']))
		$add = $_SERVER['PHP_SELF'];

	return $add . "/?vv=" . $area;
}

function rg_re_userpage($ui)
{
	if (!isset($ui['organization'])) {
		rg_internal_error("rg_re_userpage called with wrong ui (no org)!");
		rg_log("ui: " . print_r($ui, TRUE));
		exit(1);
	}

	$prefix = "";
	if ($ui['organization'] == 0)
		$prefix = "/user";

	$s = $prefix . "/" . rawurlencode($ui['username']);

	return rg_re_url($s);
}

function rg_re_repopage($ui, $repo_name)
{
	$s = rg_re_userpage($ui) . "/" . rawurlencode($repo_name);

	return rg_re_url($s);
}

function rg_re_bugpage($ui, $repo_name, $bug_id)
{
	return rg_re_repopage($ui, $repo_name) . "/bug/" . $bug_id;
}

function rg_base_url()
{
	if (!isset($_SERVER['REMOTE_ADDR']))
		return "http://SERVER/";

	$port = "";
	if (isset($_SERVER['HTTPS'])) {
		$proto = "https";
		if ($_SERVER['SERVER_PORT'] != 443)
			$port = ":" . $_SERVER['SERVER_PORT'];
	} else {
		$proto = "http";
		if ($_SERVER['SERVER_PORT'] != 80)
			$port = ":" . $_SERVER['SERVER_PORT'];
	}

	return $proto . "://" . $_SERVER['SERVER_NAME'] . $port;
}

function rg_re_repo_ssh($organization, $user, $repo)
{
	global $rg_ssh_host;
	global $rg_ssh_port;

	if ($rg_ssh_port == 22)
		$port = "";
	else
		$port = ":" . $rg_ssh_port;

	$prefix = "";
	if ($organization == 0)
		$prefix = "/user";

	return "ssh://rocketgit@" . $rg_ssh_host . $port
		. $prefix . "/" . rawurlencode($user) . "/" . rawurlencode($repo);
}

function rg_re_repo_git($organization, $user, $repo)
{
	global $rg_git_host;
	global $rg_git_port;

	if ($rg_git_port == 9418)
		$port = "";
	else
		$port = ":" . $rg_git_port;

	$prefix = "";
	if ($organization == 0)
		$prefix = "/user";

	return "git://" . $rg_git_host . $port
		. $prefix . "/" . rawurlencode($user) . "/" . rawurlencode($repo);
}

function rg_var_str($name)
{
	$ret = "";

	if (isset($_SERVER[$name]))
		$ret = $_SERVER[$name];
	else if (isset($_POST[$name]))
		$ret = $_POST[$name];
	else if (isset($_GET[$name]))
		$ret = $_GET[$name];

	return $ret;
}

function rg_var_int($name)
{
	$r = rg_var_str($name);

	if (is_array($r)) {
		$ret2 = array();
		foreach ($r as $k => $v)
			$ret2[$k] = sprintf("%d", $v);
		return $ret2;
	}
	return sprintf("%d", $r);
}

function rg_var_uint($name)
{
	$r = rg_var_str($name);

	if (is_array($r)) {
		$ret2 = array();
		foreach ($r as $k => $v)
			$ret2[$k] = sprintf("%u", $v);
		return $ret2;
	}
	return sprintf("%u", $r);
}

function rg_var_bool($name)
{
	$r = rg_var_str($name);
	if (strcmp($r, '1') == 0)
		return 1;

	return 0;
}

function rg_var_re($name, $re)
{
	$a = rg_var_str($name);
	return preg_replace($re, '', $a);
}

/*
 * Extract a cookie from $_COOKIE
 */
function rg_var_cookie_re($name, $re)
{
	if (!isset($_COOKIE[$name]))
		return "";

	return preg_replace($re, '', $_COOKIE[$name]);
}

/*
 * Enforce chars in a name. It is used for user and repo.
 */
function rg_chars_allow($name, $allowed_regexp)
{
	if (preg_match($allowed_regexp, $name) !== 1) {
		rg_log("chars_allow: [$name] does not match [$allowed_regexp]");
		return FALSE;
	}

	return TRUE;
}

/*
 * Deletes a folder and the files inside it
 */
function rg_rmdir($dir)
{
	if (!is_dir($dir)) {
		rg_util_set_error("WARN: asked to remove a non-existing dir ($dir)");
		return TRUE;
	}

	$scan = glob($dir . "/*");
	if ($scan === FALSE) {
		rg_util_set_error("invalid pattern [$dir/*]");
		return FALSE;
	}

	if (count($scan) > 0) {
		$all_good = TRUE;
		foreach ($scan as $junk => $path) {
			if (is_dir($path)) {
				$r = rg_rmdir($path);
				if ($r !== TRUE)
					return FALSE;

				continue;
			}

			if (!unlink($path)) {
				rg_util_set_error("cannot remove [$path] ($php_errormsg)");
				return FALSE;
			}
		}
	}

	if (!rmdir($dir)) {
		rg_util_set_error("cannot remove main dir ($php_errormsg)");
		return FALSE;
	}

	return TRUE;
}

/*
 * Provides a link to an image, taking in consideration the theme
 * Used by rg_prepare_image.
 */
function rg_image_callback($matches)
{
	global $rg_theme_dir;
	global $rg_theme;

	$n = $matches[1];
	$url = "/themes/" . $rg_theme . "/" . $n;
	$xfile = $rg_theme_dir . "/" . $rg_theme . "/" . $n;
	if (!is_file($xfile))
		$url = "/themes/default/" . $n;

	return $url;
}

/*
 * Prepares the images to point to a correct image.
 */
function rg_prepare_image($s)
{
	return preg_replace_callback('/@@IMG:(.*)@@/uU', "rg_image_callback", $s);
}

/*
 * Resolve variables like '@@func:func_name:paras@@'
 */
function rg_prepare_func($s, &$what, &$values)
{
	global $rg_template_functions;

	$seen = array();

	$r = preg_match_all('/@@func:(.*):(.*)@@/uU', $s, $matches, PREG_SET_ORDER);
	foreach ($matches as $i) {
		$k = '/' . $i[0] . '/';
		$func = $i[1];
		$var = $i[2];

		/* Function already called? */
		if (isset($seen[$k]))
			continue;

		if (!isset($rg_template_functions[$func]))
			continue;

		$v = preg_replace($what, $values, "@@" . $var . "@@");
		$final = $rg_template_functions[$func]($v);

		$what[] = $k;
		$values[] = $final;

		$seen[$k] = 1;
	}
}

/*
 * Helper for rg_prepare_replace
 */
function rg_prepare_replace_helper($a, $prefix, &$what, &$values)
{
	foreach ($a as $k => $v) {
		if (empty($prefix))
			$add = "";
		else
			$add = "::";

		$new_prefix = $prefix . $add . $k;

		if (is_array($v)) {
			rg_prepare_replace_helper($v, $new_prefix,
				$what, $values);
			continue;
		}

		if (strncmp($k, "HTML:", 5) == 0) {
			$k = substr($k, 5);
			$new_prefix = $prefix . $add . $k;
		} else {
			$v = rg_xss_safe($v);
		}

		$what[$new_prefix] = "/@@" . preg_quote($new_prefix, '/') . "@@/uU";
		$values[$new_prefix] = $v;
	}
}

function rg_prepare_replace(&$data, &$what, &$values)
{
	if (!empty($data)) {
		if (!is_array($data))
			rg_internal_error("prepare_replace: invalid type passed");

		rg_prepare_replace_helper($data, "", $what, $values);
	}

	$what['DUMP'] = "/@@DUMP@@/uU";
	$values['DUMP'] = rg_xss_safe(print_r($data, TRUE));

	// we replace @@unknown@@ with empty
	//$what['FINAL'] = "/@@.*@@/U";
	//$values['FINAL'] = "";

	//rg_log_ml("DEBUG: what: " . print_r($what, TRUE));
	//rg_log_ml("DEBUG: values: " . print_r($values, TRUE));
}

/*
 * Lookup a var into data array, if needed.
 * It is used for conditionals.
 */
function rg_replace_lookup(&$data, $var)
{
	rg_prepare_replace($data, $what, $values);
	return preg_replace($what, $values, $var);
}

/*
 * Helper for rg_replace_conditionals.
 * It works at line level.
 * @master_block (TRUE / FALSE) is the condition for parent block.
 */
function rg_replace_conditionals_block($block, &$data, &$stack)
{
	//rg_log("DEBUG: rg_replace_conditionals_block: block=[$block]"
	//	. " stack=" . rg_array2string($stack));

	if (!is_string($block)) {
		rg_internal_error("Block is not a string!");
		return FALSE;
	}

	// Nesting error
	if (empty($stack)) {
		rg_internal_error("Nesting error!");
		return FALSE;
	}

	$cond = array_pop($stack);
	$stack[] = $cond;
	//rg_log("cond is " . ($cond ? "TRUE" : "FALSE"));

	// First, try to match a start of 'if'
	$match1 = "@@if\s*\((.*?)\)\s*?{{";
	$match2 = "{{";
	$match3 = "}}";
	$search = $match1 . '|' . $match2 . '|' . $match3;
	$r = preg_match('/^(.*?)(' . $search . ')(.*)/su',
		$block, $matches);
	if ($r === FALSE)
		return FALSE;
	if ($r === 1) {
		//rg_log("DEBUG: matches: " . rg_array2string($matches));

		$ret = "";
		if ($cond === TRUE)
			$ret = $matches[1];

		$rest = $matches[4];
		if (strcmp($matches[2], "}}") == 0) {
			// We pop from stack only at }} and not at {{
			//rg_log("}}: Pop the stack!");
			array_pop($stack);
		} else if (strcmp($matches[2], "{{") == 0) {
			//rg_log("{{");
		} else {
			//rg_log("DEBUG: cond=" . $matches[3]);
			if (empty($matches[3])) {
				$new_cond = FALSE;
			} else {
				$r = preg_match('/^\s*(.*?)\s*(==|!=)\s*(.*?)\s*$/Dsu',
					$matches[3], $matches2);
				if ($r === FALSE) {
					rg_internal_error("Invalid condition!");
					return FALSE;
				}
				if (count($matches2) < 3) {
					rg_log("matches[3]: " . $matches[3]);
					rg_log("matches2: " . rg_array2string($matches2));
					rg_internal_error("Cannot match condition.");
					return FALSE;
				}
				$left = rg_replace_lookup($data, trim($matches2[1]));
				$op = trim($matches2[2]);
				$right = rg_replace_lookup($data, trim($matches2[3]));
				//rg_log("DEBUG: if left=[$left] op=[$op] right=[$right]");
				if (empty($op)) {
					$new_cond = empty($left) ? FALSE : TRUE;
				} else if (strcmp($op, "==") == 0) {
					$new_cond = strcmp($left, $right) == 0 ? TRUE : FALSE;
				} else if (strcmp($op, "!=") == 0) {
					$new_cond = strcmp($left, $right) == 0 ? FALSE : TRUE;
				} else {
					rg_internal_error("Invalid operation!");
					return FALSE;
				}
			}
			$not_new_cond = $new_cond ? FALSE : TRUE;

			// We have to respect the outer block condition
			// We have to push in reverse order. Remeber, first in, last out
			$stack[] = $not_new_cond && $cond;
			$stack[] = $new_cond && $cond;
		}

		$tmp = rg_replace_conditionals_block($rest, $data, $stack);
		if ($tmp === FALSE)
			return FALSE;

		//rg_log("DEBUG: returning [" . $ret . $tmp . "]");
		return $ret . $tmp;
	}

	if ($cond === FALSE)
		$block = "";

	//rg_log("DEBUG: returning [$block]");
	return $block;
}

/*
 * Replace conditionals
 * @@if(X == Y){{A}}{{B}} - if X == Y it will return A, else B.
 * Do note that there is no ending @@.
 * Because of complexity, I choosed to have a restriction: the 'if' is on
 * a single line or '@@if(...){{', '}}{{' and '}}' are on separate lines.
 * TODO: Also, we must have both branches (both true and false), for now.
 * We support nested ifs.
 */
function rg_replace_conditionals($block, &$data)
{
	rg_prof_start("replace_conditionals");
	$ret = array();

	while (1) {
		$stack = array();
		$stack[] = TRUE;
		$ret = rg_replace_conditionals_block($block, $data, $stack);
		if ($ret === FALSE)
			break;

		if (empty($stack) || ($stack[0] !== TRUE)) {
			rg_internal_error("Template nesting error!");
			$ret = FALSE;
			break;
		}

		break;
	}

	rg_prof_end("replace_conditionals");
	return $ret;
}

/*
 * Loads a file if exists, else return ""
 */
function rg_file_get_contents($f)
{
	if (!file_exists($f))
		return "";

	$c = file_get_contents($f);
	if ($c === FALSE) {
		rg_internal_error("Could not load file [$f] ($php_errormsg).");
		return "";
	}

	return $c;
}

/*
 * Merges an array (a) into another (src), using a namespace
 * Protects modifiers (HTML: etc.).
 */
function rg_array_merge($src, $namespace, $a)
{
	$ret = $src;

	if (empty($a))
		return $ret;

	if (!empty($namespace))
		$namespace .= "::";

	foreach ($a as $k => $v) {
		$t = explode(":", $k, 2);
		if (count($t) == 1)
			$ret[$namespace . $k] = $v;
		else
			$ret[$t[0] . ":" . $namespace . $t[1]] = $v;
	}

	return $ret;
}

/*
 * Builds a html output based on a template with header, footer and line
 * @data - in array of data for every out line: index 0 is line 1, index 1 is line 2...
 */
function rg_template_table($dir, &$data, $more)
{
	global $rg_theme_dir;
	global $rg_theme;

	rg_log("rg_template_table: $dir");

	$xdir = $rg_theme_dir . "/" . $rg_theme . "/" . $dir;
	if (!is_dir($xdir)) {
		rg_log("$xdir not found.");
		$xdir = $rg_theme_dir . "/default/" . $dir;
		rg_log("Using [$xdir]");
	}

	$m_what = array(); $m_values = array();
	rg_prepare_replace($more, $m_what, $m_values);

	if (!is_array($data) || empty($data)) {
		$no_data = rg_file_get_contents($xdir . "/nodata.html");
		$r = rg_replace_conditionals($no_data, $more);
		return preg_replace($m_what, $m_values, $r);
	}

	$head = rg_file_get_contents($xdir . "/header.html");
	$line = rg_file_get_contents($xdir . "/line.html");
	$foot = rg_file_get_contents($xdir . "/footer.html");
	$between = rg_file_get_contents($xdir . "/between.html");

	$head = rg_replace_conditionals($head, $more);
	$foot = rg_replace_conditionals($foot, $more);
	$between = rg_replace_conditionals($between, $more);

	$head = preg_replace($m_what, $m_values, $head);
	$foot = preg_replace($m_what, $m_values, $foot);
	$between = preg_replace($m_what, $m_values, $between);

	$body = "";
	$first = 1;
	foreach ($data as $index => $info) {
		$what = $m_what; $values = $m_values;

		rg_prepare_replace($info, $what, $values);
		rg_prepare_func($line, $what, $values);

		$line = rg_prepare_image($line);

		if ($first == 1) {
			$first = 0;
		} else {
			$body .= $between;
		}

		$more2 = array_merge($more, $info);
		$r = rg_replace_conditionals($line, $more2);
		$body .= preg_replace($what, $values, $r);
	}

	return $head . $body . $foot;
}

/*
 * Loads a template.
 * TODO: why we pass variable by reference?!
 */
function rg_template($file, &$data)
{
	global $rg_theme_dir;
	global $rg_theme;

	rg_prof_start("rg_template");
	rg_log_enter("rg_template: $file");

	$ret = "";
	while (1) {
		$xfile = $rg_theme_dir . "/" . $rg_theme . "/" . $file;
		if (!is_file($xfile)) {
			$xfile = $rg_theme_dir . "/default/" . $file;
			if (!is_file($xfile))
				$xfile = $file;
				if (!is_file($xfile)) {
					rg_log("Cannot find $file!");
					break;
				}
		}

		$body = rg_file_get_contents($xfile);
		if (empty($body)) {
			rg_log("File [$xfile] is empty. Return ''.");
			break;
		}

		$what = array();
		$values = array();

		rg_prepare_replace($data, $what, $values);
		rg_prepare_func($body, $what, $values);

		$body = rg_prepare_image($body);

		$r = rg_replace_conditionals($body, $data);
		// TODO: check error code!

		$ret = preg_replace($what, $values, $r);
		// TODO: check error code!

		break;
	}

	rg_log_exit();
	rg_prof_end("rg_template");
	return $ret;
}

/*
 * Outputs a numbered list
 */
function rg_template_list($c)
{
	$a = explode("\n", $c);
	if (count($a) == 0)
		return "";

	$ret = "";
	$i = 1;
	$add = "";
	foreach ($a as $line) {
		$ret .= $add . $i . "&nbsp;" . rg_xss_safe($line);
		$add = "<br />";
		$i++;
	}

	return $ret;
}

/*
 * Show errors using a template
 */
function rg_template_errmsg($a)
{
	if (empty($a))
		return "";

	$b = array();
	foreach ($a as $junk => $err)
		$b[] = array("error" => $err);
	return rg_template_table("errmsg", $b, array());
}

/*
 * Show a warning using a template
 */
function rg_warning($msg)
{
	if (empty($msg))
		return "";

	rg_log("Warning: $msg");

	$x = array("msg" => $msg);
	return rg_template("warning.html", $x);
}

/*
 * Show an OK message using a template
 * TODO: OBSOLETE? Because we want the files to be in templates?
 */
function rg_ok($msg)
{
	if (empty($msg))
		return "";

	$x = array("msg" => $msg);
	return rg_template("ok.html", $x);
}

/*
 * Execute $cmd and returns the output as a string, binary safe
 */
function rg_exec($cmd)
{
	rg_prof_start("exec");
	rg_log_enter("Executing [$cmd]...");

	$ret = array();
	$ret['ok'] = 0;
	$ret['errmsg'] = "";
	$ret['code'] = 65000; // fake code
	while (1) {
		$desc = array(
			0 => array("pipe", "r"),
			1 => array("pipe", "w"),
			2 => array("pipe", "w")
			);
		$a = proc_open($cmd, $desc, $pipes);
		if ($a === FALSE) {
			$ret['errmsg'] = "cannot call proc_open";
			break;
		}

		$stderr = "";
		$ret['data'] = "";
		$rx = array($pipes[1], $pipes[2]);
		while (!empty($rx)) {
			$revents = $rx;
			$wevents = NULL;
			$ex = NULL;
			$r = stream_select($revents, $wevents, $ex, 10, 0);
			if ($r === FALSE) {
				$ret['errmsg'] = "cannot select";
				break;
			}
			//rg_log("DEBUG: stream_select returned $r"
			//	. ", revents: " . rg_array2string($revents));

			foreach ($revents as $fd) {
				if (!empty($ret['errmsg']))
					break;

				//rg_log("Event on fd $fd!");
				if (feof($fd)) {
					//rg_log("eof on fd $fd!");
					foreach ($rx as $_key => $_fd)
						if ($_fd == $fd)
							unset($rx[$_key]);
					continue;
				}

				$_d = fread($fd, 32 * 4096);
				if ($_d === FALSE) {
					$ret['errmsg'] = "cannot read";
					break;
				}

				//rg_log("DEBUG: got data from fd $fd [$_d]");

				if ($fd === $pipes[2])
					$stderr .= $_d;
				else if ($fd === $pipes[1])
					$ret['data'] .= $_d;
			}

			if (!empty($ret['errmsg']))
				break;
		}
		$stderr = trim($stderr);
		$ret['data'] = trim($ret['data']);

		for ($i = 0; $i < 3; $i++)
			fclose($pipes[$i]);

		$err = proc_close($a);
		if ($err != 0) {
			$ret['code'] = $err;
			$ret['errmsg'] = "code " . $err . ": " . $stderr;
		}

		if (empty($ret['errmsg'])) {
			$ret['code'] = 0;
			$ret['ok'] = 1;
		}

		break;
	}

	rg_log_exit();
	rg_prof_end("exec");
	return $ret;
}

/*
 * Force browser to redirect to another page
 */
function rg_redirect($url)
{
	rg_log("Redirecting to [$url]");
	header("Location: $url");
	exit(0);
}

/*
 * Force browser to redirect to another page, using a HTML header
 */
function rg_redirect_html($seconds, $url)
{
	global $rg;

	$rg['rg_redirect_html'] = 1;
	$rg['rg_redirect_html_seconds'] = $seconds;
	$rg['rg_redirect_html_url'] = $url;
}

/*
 * Transforms strange chars to hexa
 */
function rg_callback_hexa($matches)
{
	$n = pack("a*", $matches[0]);
	$tmp = unpack("H*", $n);
	return "[" . $tmp[1] . "]";
}

/*
 * Transforms an array in a string
 */
function rg_array2string($a)
{
	$what = array("/[^\pL\pN\pP\pS ]/uU");

	if (!is_array($a))
		return preg_replace_callback($what, "rg_callback_hexa", $a);

	if (empty($a))
		return "";

	$ret = "";
	$add = "";
	foreach ($a as $k => $v) {
		if (is_array($v))
			$s = rg_array2string($v);
		else
			$s = preg_replace_callback($what, "rg_callback_hexa", $v);
		$ret .= $add . "$k=[$s]";
		$add = " ";
	}

	return $ret;
}

/*
 * Load files from a folder using a pattern for match
 */
function rg_dir_load_pattern($dir, $pattern)
{
	$ret = FALSE;
	if (!file_exists($dir)) {
		rg_util_set_error("$dir does not exists");
		return $ret;
	}

	$d = @scandir($dir);
	if ($d === FALSE) {
		rg_util_set_error("cannot scan dir $dir ($php_errormsg)");
		return $ret;
	}

	$ret = array();
	foreach ($d as $file) {
		if ((strcmp($file, ".") == 0) || (strcmp($file, "..") == 0))
			continue;

		if (preg_match("/" . $pattern . "/", $file) !== 1)
			continue;

		$ret[] = $file;
	}

	return $ret;
}

/*
 * Load all files from a folder
 */
function rg_dir_load($dir)
{
	return rg_dir_load_pattern($dir, ".*");
}

/*
 * Recursive dir load (used for references)
 */
function rg_dir_load_deep($dir)
{
	$ret = array();

	if (is_file($dir))
		return array($dir);

	$d = rg_dir_load($dir);
	foreach ($d as $obj) {
		if (is_dir($dir . "/" . $obj)) {
			$c = rg_dir_load_deep($dir . "/" . $obj);
			foreach ($c as $obj2)
				$ret[] = $obj . "/" . $obj2;
		} else {
			$ret[] = $obj;
		}
	}

	return $ret;
}

/*
 * Copy a fs tree to another place
 */
function rg_copy_tree($src, $dst, $mask)
{
	rg_log("rg_copy_tree($src, $dst, $mask)");

	if (!is_dir($dst)) {
		$r = @mkdir($dst, $mask);
		if ($r !== TRUE) {
			rg_log("ERROR: Cannot mkdir [$dst] ($php_errormsg).");
			return FALSE;
		}
	}

	$d = rg_dir_load($src);
	foreach ($d as $obj) {
		if (is_dir($src . "/" . $obj)) {
			if (!is_dir($dst . "/" . $obj)) {
				$r = @mkdir($dst . "/" . $obj, $mask);
				if ($r !== TRUE) {
					rg_log("ERROR: Cannot mkdir [$dst/$obj]"
						. " ($php_errormsg).");
					return FALSE;
				}
			}

			$r = rg_copy_tree($src . "/" . $obj, $dst . "/" . $obj, $mask);
			if ($r !== TRUE)
				return FALSE;
		} else {
			$r = @copy($src . "/" . $obj, $dst . "/" . $obj);
			if ($r !== TRUE) {
				rg_log("ERROR: Cannot copy file ($php_errormsg).");
				return FALSE;
			}
		}
	}

	return TRUE;
}

/*
 * Called by PHP in case of error
 */
function rg_error_handler($no, $str, $file, $line)
{
	if ($no == 0)
		return;

	// call was prepended with '@'
	if (error_reporting() == 0)
		return;

	$msg = "PHP ERROR: $file:$line: $str (errno=$no)";
	$key = md5($msg);
	rg_error_core($msg);

	if ($no == E_ERROR)
		die();

	$rg_error_seen[$key] = 1;

	return FALSE;
}

/*
 * Shutdown function to log fatal errors
 */
function rg_error_shutdown()
{
	$a = error_get_last();
	if ($a === NULL)
		return;

	rg_error_handler($a['type'], $a['message'], $a['file'], $a['line']);
}

/*
 * YYYY-MM-DD -> timestamp
 */
function rg_date2ts($s)
{
	rg_log("rg_date2ts s=[$s]");

	if (strlen($s) != 10)
		return FALSE;

	$f = explode("-", $s);

	if (count($f) != 3)
		return FALSE;

	return gmmktime(0, 0, 0, $f[1], $f[2], $f[0]);
}

/*
 * YYYY-MM-DD -> timestamp (last second of the day)
 */
function rg_date2ts_last_second($s)
{
	rg_log("rg_date2ts_last_second s=[$s]");

	if (strlen($s) != 10)
		return FALSE;

	$f = explode("-", $s);

	if (count($f) != 3)
		return FALSE;

	return gmmktime(0, 0, 0, $f[1], $f[2] + 1, $f[0]) - 1;
}

/*
 * Function to send e-mails
 * TODO: Replace mail() with rg_mail everywhere.
 */
function rg_mail_template($template, $more)
{
	global $rg_admin_name, $rg_admin_email;

	rg_prof_start("mail_template");
	rg_log("mail_template: $template, more=" . rg_array2string($more));

	// Account was not confirmed, so do not send mail
	if (empty($more['ui::email']))
		return TRUE;

	$more['HTML:rg_admin_email'] = $rg_admin_email;
	$more['HTML:utf8_rg_admin_name'] = "=?UTF-8?B?"
		. base64_encode($rg_admin_name) . "?=";

	$subject = rg_template($template . ".subj.txt", $more);
	$subject = "=?UTF-8?B?" . base64_encode(trim($subject)) . "?=";
	$header = rg_template("mail/common.head.txt", $more);
	$header .= rg_template($template . ".head.txt", $more);
	$header = trim($header);
	$body = rg_template($template . ".body.txt", $more);

	rg_log("CHECK: mail_template(" . $more['ui::email'] . ",
		$subject, $body, $header, -f $rg_admin_email");
	$ret = mail($more['ui::email'], $subject, $body, $header,
		"-f $rg_admin_email");
	if ($ret === FALSE)
		rg_log("Sending mail failed to=" . $more['ui::email'] . " subject=$subject!");

	rg_prof_end("mail_template");
	return $ret;
}

/*
 * Special implode, with prefix/postfix
 */
function rg_implode($prefix, $a, $postfix)
{
	if (!is_array($a))
		return $a;

	if (empty($a))
		return "";

	$ret = array();
	foreach ($a as $index => $data)
		$ret[] = $prefix . $data;

	return implode($postfix, $ret);
}

/*
 * Here we will cache the connections
 */
$rg_socket_cache = array();

/*
 * Receives buffers and test if @wait string is present.
 * @timeout - in miliseconds, NULL=forever, 0=no_wait
 */
function rg_socket_recv_wait($socket, $wait, $timeout)
{
	$ret = FALSE;

	if ($timeout === NULL) {
		$tv_sec = NULL;
		$tv_usec = NULL;
	} else {
		$tv_sec = $timeout / 1000;
		$tv_usec = ($timeout % 1000) * 1000;
	}

	$ret_buf = "";
	while (1) {
		$reads = array($socket); $writes = array(); $ex = array();
		$r = @socket_select($reads, $writes, $ex, $tv_sec, $tv_usec);
		if ($r === FALSE) {
			rg_log("Cannot select(" . socket_strerror(socket_last_error()) . ")!");
			break;
		}

		if ($r === 0) { // timeout
			rg_log("Timeout in reading!");
			break;
		}

		if (!in_array($socket, $reads)) {
			rg_log("Select returned > 0 and my socket is not in reads");
			break;
		}

		$r = @socket_recv($socket, $buf, 32 * 4096, 0);
		if ($r === FALSE) {
			rg_log("Cannot receive(" . socket_strerror(socket_last_error()) . ")!");
			break;
		}
		//rg_log("Received [$buf]");
		$ret_buf .= $buf;

		$pos = strpos($buf, $wait);
		if ($pos !== FALSE) {
			$ret = $ret_buf;
			break;
		}

		break;
	}

	return $ret;
}

/*
 * Sends a full buffer
 * TODO: Take timeout in consideration.
 */
function rg_socket_send($socket, $buf)
{
	$ret = FALSE;
	$len = strlen($buf);
	$off = 0;
	while (1) {
		$r = @socket_send($socket, substr($buf, $off), $len - $off, 0);
		if ($r === FALSE) {
			rg_log("Could not send (" . socket_strerror(socket_last_error()) . ")!");
			break;
		}
		//rg_log("Sent $r bytes (" . substr($buf, $off, $r) . ").");

		$len -= $r; $off += $r;
		if ($len == 0) {
			$ret = TRUE;
			break;
		}
	}

	return $ret;
}

/*
 * Connects to a socket, send @buf and returns the answer.
 * @timeout: NULL=forever, 0=no_wait
  is 0, we do not wait for an answer. If is NULL, we wait forever.
 * @tries - how many time to retry if it fails
 */
function rg_socket($path, $buf, $timeout, $tries)
{
	global $rg_socket_cache;

	rg_prof_start("socket");

	$ret = FALSE;
	while ($tries > 0) {
		if (isset($rg_socket_cache[$path])) {
			$socket = $rg_socket_cache[$path];
		} else {
			$socket = @socket_create(AF_UNIX, SOCK_STREAM, 0);
			if ($socket === FALSE) {
				rg_log("Could not create socket (" . socket_strerror(socket_last_error()) . ")!");
				break;
			}

			while ($tries > 0) {
				$r = @socket_connect($socket, $path);
				if ($r === FALSE) {
					$tries--;
					usleep(50 * 1000);
					continue;
				}

				break;
			}
			if ($r === FALSE) {
				rg_log("Could not connect the socket (" . socket_strerror(socket_last_error()) . ")!");
				break;
			}

			$rg_socket_cache[$path] = $socket;
		}

		$r = rg_socket_send($socket, $buf);
		if ($r !== TRUE) {
			socket_close($socket);
			unset($rg_socket_cache[$path]);
			continue;
		}

		if ($timeout === 0) {
			rg_log("We do not have to wait. Exit.");
			$ret = "";
			break;
		}

		$ret = rg_socket_recv_wait($socket, "\n", $timeout);
		break;
	}

	rg_prof_end("socket");
	return $ret;
}

/*
 * Check if referer matchces current website
 */
function rg_valid_referer()
{
	$ref0 = isset($_SERVER['HTTP_REFERER']) ? $_SERVER['HTTP_REFERER'] : "";

	// If not provided, we can do nothing about
	if (empty($ref0))
		return TRUE;

	$ref = preg_replace('|http(s)?://|', '', $ref0);
	$ref = preg_replace('|/.*|', '', $ref);

	$we = isset($_SERVER['HTTP_HOST']) ? $_SERVER['HTTP_HOST'] : "";

	if (strcasecmp($we, $ref) == 0)
		return TRUE;

	rg_security_violation_no_exit("invalid referer for form submission"
		. " we=[$we] ref=[$ref] ref0=[$ref0]");
	return FALSE;
}

?>
